/*
** prims.c
** march 1994 by Maarten Pennings
*/

#include <stdio.h>

#include "data_layout.h"

#include "ints.h"
#include "bools.h"
#include "chars.h"
#include "reals.h"
#include "strs.h"
#include "toks.h"
#include "ptrs.h"
#include "maps.h"
#include "attrs.h"

#include "prims.h"
#include "memo.h"

#include "visit_tables.h"
#include "grammar_tables.h"

void lrc_lib_init() {
  lrc_term_init(); 
  lrc_Int_init();
  lrc_Bool_init();
  lrc_Char_init();
  lrc_Str_init();
  lrc_memo_init();
  lrc_dump_reset();
  return;
}

void lrc_lib_mark() {
  lrc_Int_mark();
  lrc_Bool_mark();
  lrc_Char_mark();
  return;
}

/* The ssl-expression (l1 @ l2) is translated into
   the C function-call
      lrc_list_append(T_Prod0,l1,l2)
   (when l1 and l1 have type L:
     list L;
     L : Prod0() | Prod2(Element_type L);
   )
 */
lrc_term lrc_list_append(int empty_tag, lrc_term l1, lrc_term l2) {
  if (LRC_TAG(l1)==empty_tag) return l2; 
  if (LRC_TAG(l2)==empty_tag) return l1;
  /* LRC_TAG(l1)==LRC_TAG(l2)==the tag of the binary prod (T_Prod2 in the example) */
  return lrc_term2(LRC_TAG(l1),LRC_CHILD(l1,1)
             ,lrc_list_append(empty_tag,LRC_CHILD(l1,2),l2)
             );
}

void lrc_dump_reset() {
  lrc_term_dump_set_birth_time(lrc_term_time());  
  lrc_memo_dump_set_birth_time(lrc_memo_time());  
  /* Would you like to see all: */
  /* lrc_term_dump_set_birth_time(0); */
  /* lrc_memo_dump_set_birth_time(0); */
  return;
}

lrc_bool lrc_filter_true(lrc_term t) {
  return lrc_true;
}


/*******************\
** current install **
\*******************/

static  lrc_prim_print_function  prim_printer = lrc_prim_print_untagged;
static  lrc_term_print_function  term_printer = lrc_term_print_long;
static  lrc_cell_print_function  cell_printer = lrc_cell_print_fipop;


#define PROD_TAG_NAME(ptag) (lrc_constructor_ident(ptag))
#define TAG_NAME(t)   PROD_TAG_NAME(LRC_TAG(t))
#define NUM_SONS(t)   (lrc_constructor_num_sons(LRC_TAG(t)))
#define VPR_IX(f)     lrc_vpr_index(f)
#define VPR_ROUTINE(i)  lrc_vpr_routine(i)
#define VPR_PROD_NUMBER(i) lrc_vpr_prod_number(i)
#define VPR_PROD_NAME(i) PROD_TAG_NAME(VPR_PROD_NUMBER(i))
#define VPR_VISIT_NUMBER(i) lrc_vpr_visit_number(i)
#define VPR_NUM_IP(i) lrc_vpr_num_ip_attrs(i)
#define VPR_NUM_OP(i) lrc_vpr_num_op_attrs(i)


/* compare two terms.
 *
 */

int LrcCompareTrees(lrc_term t0,lrc_term t1) {
     if (t0==t1) { return lrc_equal;}
else if (LRC_TAG(t0)<LRC_TAG(t1)) { return lrc_less;}
else if (LRC_TAG(t0)>LRC_TAG(t1)) { return lrc_greater;}
else {
   if (PRIMITIVE(t0)) {
     return lrc_primitive_compare(t0,t1);
   }
   else {
     int num_sons=lrc_constructor_num_sons(LRC_TAG(t0));
     int son_num;
     for(son_num=1;son_num<=num_sons;son_num++) {
       int son_compare=LrcCompareTrees(LRC_CHILD(t0,son_num),LRC_CHILD(t1,son_num));
       if (son_compare!=lrc_equal) return son_compare;
     }
     return lrc_equal;
   }
 }
}

int lrc_term_equal(lrc_term t0, lrc_term t1) {
  lrc_term s0,s1;
  
  if (t0==0) return t1==0; /* Hack: should not happen */
  if (t1==0) return t0==0; /* Hack: should not happen */
  s0=(INITIAL(t0)?SHADOW(t0):t0);
  s1=(INITIAL(t1)?SHADOW(t1):t1);
  return s0==s1;
}

/*******************\
** prim comparison **
\*******************/
/* Compare two primitive values.
   Terms t0 and t1 have the same tag.
 */

int lrc_primitive_compare(lrc_term t0, lrc_term t1) {
   switch (LRC_TAG(t0)) {
     case T__Int  : return lrc_int_compare_term(t0,t1);
     case T__Bool : return lrc_bool_compare_term(t0,t1);
     case T__Char : return lrc_char_compare_term(t0,t1);
     case T__Real : return lrc_real_compare_term(t0,t1);
     case T__Tok  : /* fall through */
     case T__Str  : return lrc_string_compare_term(t0,t1);
     case T__Ptr  : return lrc_ptr_compare_term(t0,t1);
     case T__Map  : return lrc_map_compare_term(t0,t1);
     case T__Attr : return lrc_attr_compare_term(t0,t1);
     default : printf("lrc_primitive_compare called with wrong term. Tag=%d\n", LRC_TAG(t0));
               return lrc_less;
   }
}

/*****************\
** prim printing **
\*****************/

void lrc_prim_print_address(lrc_term t) {
  if (PRIMITIVE(t)) printf("@%p",t);
    else printf("UNKNOWN in lrc_prim_print_address");
}

void lrc_prim_fprint_untagged(FILE * f, lrc_term t) {
  switch (LRC_TAG(t)) {
    case T__Int:
      fprintf(f,"%d",LRC_VALUE(t,int));
      break;
    case T__Bool:
      fprintf(f,LRC_VALUE(t,lrc_bool) ? "true" : "false");
      break;
    case T__Char:
      fprintf(f,"'"); lrc_char_fprint(f,LRC_VALUE(t,char)); fprintf(f,"'");
      break;
    case T__Real:
      fprintf(f,"%f",LRC_VALUE(t,lrc_real));
      break;
    case T__Str:
      fprintf(f,"\""); lrc_string_fprint(f,LRC_VALUE(t,lrc_string)); fprintf(f,"\"");
      break;
    case T__Tok:
      fprintf(f,"<Tok not implemented>");
      break;
    case T__Ptr:
      fprintf(f,"&%p",LRC_VALUE(t,lrc_ptr));
      break;
    case T__Map:
      /* TODO: */ lrc_map_print(LRC_VALUE(t,lrc_map),term_printer,term_printer);
      break;
    case T__Attr:
      fprintf(f,"&&%p",LRC_VALUE(t,lrc_attr));
      break;
    default: 
      fprintf(f,"UNKNOWN");
  }
}

void lrc_prim_print_untagged(lrc_term t) {
   lrc_prim_fprint_untagged(stdout,t);
}

void lrc_prim_print_short(lrc_term t) {
  switch (LRC_TAG(t)) {
    case T__Str:
      printf("<string>");
      break;
    case T__Map:
      printf("<map>");
      break;
    default: 
      lrc_prim_print_untagged(t);
  }
}

void lrc_prim_print_tagged(lrc_term t) {
  printf("%s(",TAG_NAME(t));
  lrc_prim_print_untagged(t); printf(")");
}

void lrc_prim_print_install(lrc_prim_print_function p) {
  prim_printer=p;
  return;
}



/*****************\
** term printing **
\*****************/

void lrc_term_print_address(lrc_term t) {
  if (PRIMITIVE(t)) lrc_prim_print_address(t); 
    else printf("@%p",t);
}

void lrc_term_print_short(lrc_term t) {
  printf("@%p ",t);  
  if (PRIMITIVE(t)) {
    prim_printer(t); 
  } else { int ix;
    printf("%s(",TAG_NAME(t));
    for(ix=1; ix<=NUM_SONS(t); ix++) {
      if (ix>1) printf(",");
      printf("@%p",LRC_CHILD(t,ix));
    }
    printf(")");
  }
}

void lrc_term_print_long(lrc_term t) {
  if (t==0) { printf("<null>"); return; }
  if (LRC_TAG(t)<0) { printf("<negative tag>"); return; }
  if (LRC_TAG(t)>1000) { printf("<large tag>"); return; }

  if (PRIMITIVE(t)) {
    prim_printer(t); 
  } else { int ix;
    printf("%s(",TAG_NAME(t)); fflush(stdout);
    for(ix=1; ix<=NUM_SONS(t); ix++) {
      if (ix>1) printf(",");
      term_printer(LRC_CHILD(t,ix));
    }
    printf(")");
  }
}

static void emit_indent(FILE * f, int n) {
  int ix;
  for(ix=1;ix<=n;ix++) fprintf(f," ");
}

static int indent_delta;
static void loc_lrc_term_fprint_indent(FILE *f, lrc_term t, int indent) {
  emit_indent(f,indent);
  if (t==0) { fprintf(f,"<null>"); return; }
  if (LRC_TAG(t)<0) { fprintf(f,"<negative tag>"); return; }
  if (LRC_TAG(t)>1000) { fprintf(f,"<large tag>"); return; }

  if (PRIMITIVE(t)) {
    lrc_prim_fprint_untagged(f,t); 
  } else { 
    int ix;
    int num_children;

    fprintf(f,"%s",TAG_NAME(t)); fflush(f);
    num_children=NUM_SONS(t);

    if (num_children==0) {  /* No children */
      fprintf(f,"()");
    }
    else if ( num_children==1 && PRIMITIVE(LRC_CHILD(t,1)) ) { /* One primitive child */
      fprintf(f,"(");
      lrc_prim_fprint_untagged(f,LRC_CHILD(t,1));
      fprintf(f,")");
    }
    else { /* More than one child or a non-primitive child */
      fprintf(f,"(\n");
      for(ix=1; ix<=num_children; ix++) {
        if (ix>1) fprintf(f,",\n");
        loc_lrc_term_fprint_indent(f,LRC_CHILD(t,ix),indent+indent_delta);
      }
      fprintf(f,"\n");
      emit_indent(f,indent); 
      fprintf(f,")");
    }
  }
}

void lrc_term_print_indent(lrc_term t, int usr_indent_delta) {
  indent_delta=usr_indent_delta;
  loc_lrc_term_fprint_indent(stdout,t,0);
}

void lrc_term_fprint_indent(FILE * f, lrc_term t, int usr_indent_delta) {
  indent_delta=usr_indent_delta;
  loc_lrc_term_fprint_indent(f,t,0);
}

void lrc_term_print_install(lrc_term_print_function p) {
  term_printer=p;
  return;
}




/*****************\
** cell printing **
\*****************/

void lrc_cell_print_fipop(lrc_cell internal_cell) {
  int ix;
  int vpr;

  lrc_cell c=internal_cell+CELL_ADMIN_SIZE;
  vpr=VPR_IX(*c++);
  printf("Print fipop; vpr=%d\n",vpr); fflush(stdout);
  printf("%s\\%d(",VPR_PROD_NAME(vpr),VPR_VISIT_NUMBER(vpr)); fflush(stdout);
  term_printer((lrc_term)*c); c++;
  for(ix=0; ix<VPR_NUM_IP(vpr); ix++) {
    printf("IP[%d]",ix); fflush(stdout);
    printf(",");  fflush(stdout);term_printer((lrc_term)*c); c++; fflush(stdout);
  }
  printf(")=("); fflush(stdout);
  for(ix=0; ix<VPR_NUM_OP(vpr); ix++) {
    printf("OP[%d]",ix); fflush(stdout);
    if (ix>0) printf(","); fflush(stdout);
    term_printer((lrc_term)*c); c++;
  }
  printf(")");
  return;
}

void lrc_cell_print_install(lrc_cell_print_function p) {
  cell_printer = p;
  return;
}


